import sys
import msvcrt


def input(draw, print, pos, max_len=60, f_col=(255, 255, 255), b_col=(0, 0, 0)):

	while msvcrt.kbhit():
		msvcrt.getch()

	cursor = 0
	word = []
	user_input = '%c' % msvcrt.getwch()

	while user_input != '\r':

		if user_input == '':

			if len(word) > 0:

				cursor -= 1
				print((pos[0] + cursor, pos[1]), ' ')
				word.pop()

		else:

			if len(word) <= max_len:

				print((pos[0] + cursor, pos[1]), user_input, f_col=f_col, b_col=b_col)
				cursor += 1
				word.append(user_input)

		draw()
		user_input = '%c' % msvcrt.getwch()

	return ''.join(word)


def print(buffer, state, pos, word, f_col=(255, 255, 255), b_col=(0, 0, 0)):

	if pos[1] % 2 != 0:
		pos = (pos[0], pos[1] - 1)

	buffer[pos[1]][pos[0]] = '\x1b[48;2;%d;%d;%dm\x1b[38;2;%d;%d;%dm%s' % (*b_col, *f_col, word[0])
	state[pos[1]][pos[0]] = True

	for cursor in range(1, len(word)):

		buffer[pos[1]][pos[0] + cursor] = word[cursor]
		state[pos[1]][pos[0] + cursor] = True


import ctypes
import sys
import os

from ctypes import POINTER, WinDLL, Structure, sizeof, c_long, pointer
from ctypes.wintypes import BOOL, SHORT, WCHAR, UINT, ULONG, DWORD, HANDLE, RECT

class Terminal:

	def __init__(self):

		self.kernel32 = ctypes.windll.kernel32
		self.user32 = ctypes.windll.user32

		self.hWnd = self.kernel32.GetConsoleWindow()

		self.kernel32.SetConsoleMode(self.kernel32.GetStdHandle(-11), 7)
		self.kernel32.SetConsoleMode(self.kernel32.GetStdHandle(-10), (0x4|0x80|0x20|0x2|0x10|0x1|0x00|0x100))

		sys.stdout.write('\x1b[?25l')
		sys.stdout.flush()

		self._p_size = 8


	@property
	def name(self):
		length = self.user32.GetWindowTextLengthW(self.hWnd) + 1
		title = ctypes.create_unicode_buffer(length)
		self.user32.GetWindowTextW(self.hWnd, title, length)
		return title.value

	@name.setter
	def name(self, new_name):
		sys.stdout.write(f'\33]0;{new_name}\a')
		sys.stdout.flush()


	@property
	def size(self):
		width, height = os.get_terminal_size()
		return (width, 2*height)

	@size.setter
	def size(self, new_size):
		os.system('mode con: cols=%d lines=%d' % (new_size))


	@property
	def client_size(self):
		rect = RECT()
		self.user32.GetClientRect(self.hWnd, pointer(rect))
		return (rect.right - rect.left, rect.bottom - rect.top)

	@client_size.setter
	def client_size(self, new_size):
		width, height = new_size[0] + 32, new_size[1] + 38
		self.user32.MoveWindow(self.hWnd, *self.pos, width, height, True)


	@property
	def pos(self):
		rect = RECT()
		self.user32.GetWindowRect(self.hWnd, pointer(rect))
		return (rect.left, rect.top)
	
	@pos.setter
	def pos(self, new_pos):
		rect = RECT()
		self.user32.GetWindowRect(self.hWnd, pointer(rect))
		self.user32.MoveWindow(self.hWnd, *new_pos, rect.right - rect.left, rect.bottom - rect.top, True)


	@property
	def m_pos_client(self):
		pt = POINT()
		self.user32.GetCursorPos(pointer(pt))
		self.user32.ScreenToClient(self.hWnd, pointer(pt))
		return (pt.x, pt.y)

	@m_pos_client.setter
	def m_pos_client(self, new_pos):
		self.user32.SetCursorPos(*new_pos)


	@property
	def m_pos(self):
		px, py = round(self.m_pos_client[0]/self.client_size[0]*self.size[0]), round(self.m_pos_client[1]/self.client_size[1]*self.size[1])

		if px < 0: px = 0
		elif px > self.size[0]: px = self.size[0]
		if py < 0: py = 0
		elif py > self.size[1]: py = self.size[1]
		
		return (px, min(self.size[1], py))

	@m_pos.setter
	def m_pos(self, new_pos):
		self.m_pos_client = new_pos


	@property
	def p_size(self):
		return self._p_size
	

	@p_size.setter
	def p_size(self, size):

		font = CONSOLE_FONT_INFOEX()
		font.cbSize = sizeof(CONSOLE_FONT_INFOEX)
		font.FaceName = 'Consolas'
		font.dwFontSize.Y, font.dwFontSize.X = 2 * size, size
		font.FontFamily = 0
		self.kernel32.SetCurrentConsoleFontEx(self.kernel32.GetStdHandle(-11), c_long(False), pointer(font))
	
		self._p_size = size


	def key_pressed(self, key):
		return self.user32.GetKeyState(key) & 0x8000 and self.hWnd == self.user32.GetForegroundWindow()


	def m_click(self, key):

		if key == 'left':
			code = 0x01

		elif key == 'right':
			code = 0x02

		return self.key_pressed(code)


class POINT(Structure):
    _fields_ = [("x", c_long), ("y", c_long)]
	

class COORD(Structure):
    _fields_ = [("X", SHORT), ("Y", SHORT),]


class CONSOLE_FONT_INFOEX(Structure):
    _fields_ = [
        ("cbSize", ULONG),
        ("nFont", DWORD),
        ("dwFontSize", COORD),
        ("FontFamily", UINT),
        ("FontWeight", UINT),
        ("FaceName", WCHAR * 32)
    ]



# initialization
import nsii
nsii = nsii.Nsii()

# retrieve values of a parameter
values = nsii.parameter

# set values of a parameter
nsii.parameter = values

# available parameters
nsii.name         # window name  ->  str
nsii.size         # window size  ->  (int, int)
nsii.pos          # window position  ->  (int, int)

nsii.fps          # frame rate  ->  float
nsii.fps_target   # frame rate target  ->  int or 'max'

nsii.m_pos        # mouse coordinates  ->  (int, int)
nsii.p_size       # pixel size  -> int

# available functions
nsii.draw()  # displays all changes

nsii.m_click('left')   # mouse button status  ->  bool
nsii.m_click('right')

nsii.input((x, y))        # retrieve user input  ->  str
nsii.print((x, y), text)  # same as print()

nsii.key_pressed(key)  # retrieve the status of the given key code  ->  bool
					   # for key code see https://docs.microsoft.com/en-us/windows/win32/inputdev/virtual-key-codes

nsii.dot(x, y)                  # draw dot
nsii.line(x0, y0, x1, y1)       # draw line
nsii.circle(x, y, radius)       # draw circle
nsii.rect(x, y, width, height)  # drax rectangle

# note that you can use a custom color by entering these parameters
f_col=(r, g, b)  # front color
b_col=(r, g, b)  # back color

# image initialisation
image = nsii.new_image('example.ppm')

# available parameter
image.size  # image size  ->  (int, int)
image.pos   # image position  ->  (int, int)

# available functions
image.show()  # displays the image

# note that you can hide a specific color by entering this parameter.
hide=(255, 255, 255)


from . import core
from . import terminal
from . import fps

from . import io
from . import image
from . import tools

import pathlib

class Nsii:

	def __init__(self):

		self._core = core.Core()
		self._terminal = terminal.Terminal()
		self._fps_handler = fps.Fps_handler()

		self._script_location = pathlib.Path(__file__).absolute().parent.parent


	@property
	def name(self):
		return self._terminal.name

	@name.setter
	def name(self, new_name):
		self._terminal.name = new_name


	@property
	def size(self):
		return self._terminal.size

	@size.setter
	def size(self, new_size):
		self._terminal.size = new_size 


	@property
	def client_size(self):
		return self._terminal.client_size

	@client_size.setter
	def client_size(self, new_size):
		self._terminal.client_size = new_size


	@property
	def pos(self):
		return self._terminal.pos

	@pos.setter
	def pos(self, new_pos):
		self._terminal.pos = new_pos


	@property
	def fps(self):
		return self._fps_handler.fps


	@property
	def fps_target(self):
		return self._fps_handler.fps_target

	@fps_target.setter
	def fps_target(self, new_target):
		self._fps_handler.fps_target = new_target


	@property
	def p_size(self):
		return self._terminal.p_size

	@p_size.setter
	def p_size(self, size):
		self._terminal.p_size = size


	@property
	def m_pos(self):
		return self._terminal.m_pos
	
	@m_pos.setter
	def m_pos(self, new_pos):
		self._terminal.m_pos = new_pos


	def draw(self, *args, **kwargs):
		self._core.draw(self._fps_handler.update, *self.size, *args, **kwargs)


	def new_image(self, path, *args, **kwargs):
		return image.Image(self._core.buffer, self._core.state, self._script_location / f'{path}', *args, **kwargs)


	def m_click(self, *args, **kwargs):
		return self._terminal.m_click(*args, **kwargs)


	def key_pressed(self, *args, **kwargs):
		return self._terminal.key_pressed(*args, **kwargs)


	def input(self, *args, **kwargs):
		return io.input(self.draw, self.print, *args, **kwargs)


	def print(self, *args, **kwargs):
		io.print(self._core.buffer, self._core.state, *args, **kwargs)


	def dot(self, *args, **kwargs):
		tools.dot(self._core.buffer, self._core.state, *args, **kwargs)


	def line(self, *args, **kwargs):
		tools.line(self._core.buffer, self._core.state, *args, **kwargs)


	def rect(self, *args, **kwargs):
		tools.rect(self._core.buffer, self._core.state, *args, **kwargs)


	def circle(self, *args, **kwargs):
		tools.circle(self._core.buffer, self._core.state, *args, **kwargs)


import time

class Fps_handler:

	def __init__(self):

		self.last_frame = time.perf_counter()
		self.last_period = time.perf_counter()

		self.time_per_frame = None
		self.fps_target = 60
		self.counter = 1


	@property
	def fps(self):
		fps = self.counter / (time.perf_counter() - self.last_period)
		self.counter = 0
		self.last_period = time.perf_counter()
		return fps


	@property
	def fps_target(self):
		return self._fps_target
	
	@fps_target.setter
	def fps_target(self, new_target):
		if new_target == 'max':
			self.time_per_frame = 0
		else:
			self.time_per_frame = 1 / new_target
		self._fps_target = new_target


	def update(self):

		next_frame = self.last_frame + self.time_per_frame

		while time.perf_counter() <= next_frame:
			pass

		self.last_frame = time.perf_counter()
		self.counter += 1


import sys

class Core:

	def __init__(self):

		self.buffer = [['8;2;0;0;0m' for a in range(2000)] for b in range(2000)]
		self.state = [[True for a in range(2000)] for b in range(2000)]


	def draw(self, update, width, height, force=False):

		x, y = 0, 0
		serie = False

		for y in range(0, height, 2):
			for x in range(width):

				if self.state[y][x] or self.state[y+1][x] or force:

					if not(serie):
						sys.stdout.write('\x1b[%d;%dH'%(y//2+1, x+1))
						serie = True

					if self.buffer[y][x][0] == '8':
						sys.stdout.write(f'\x1b[4{self.buffer[y][x]}\x1b[3{self.buffer[y+1][x]}▄')

					elif self.buffer[y][x][0] != 'None':
						sys.stdout.write(self.buffer[y][x])

					self.state[y][x] = False
					self.state[y+1][x] = False

				else:
					if serie: serie = False

		update()
		sys.stdout.flush()


def dot(buffer, state, x, y, f_col=(255, 255, 0)):

	buffer[y][x] = '8;2;%d;%d;%dm' % f_col
	state[y][x] = True


def line(buffer, state, x0, y0, x1, y1, f_col=(255, 255, 255)):

	if x0 == x1:

		for y in range(y0, y1 + 1):

			buffer[y][x0] = '8;2;%d;%d;%dm' % f_col
			state[y][x0] = True

	elif y0 == y1:

		for x in range(x0, x1 + 1):

			buffer[y0][x] = '8;2;%d;%d;%dm' % f_col
			state[y0][x] = True

	else:

		if abs(y1 - y0) < abs(x1 - x0):

			if x0 > x1:
				line_low(buffer, state, x1, y1, x0, y0, f_col)

			else:
				line_low(buffer, state, x0, y0, x1, y1, f_col)

		else:

			if y0 > y1:
				line_high(buffer, state, x1, y1, x0, y0, f_col)

			else:
				line_high(buffer, state, x0, y0, x1, y1, f_col)


def line_high(buffer, state, x0, y0, x1, y1, f_col):

	dx, dy, xi = x1 - x0, y1 - y0, 1

	if dx < 0:
		xi, dx = -1, -dx

	D, x = (2 * dx) - dy, x0

	for y in range(y0, y1 + 1):

		buffer[y][x] = '8;2;%d;%d;%dm' % f_col
		state[y][x] = True

		if D > 0:
			x = x + xi
			D = D + (2 * (dx - dy))

		else:
			D = D + 2*dx


def line_low(buffer, state, x0, y0, x1, y1, f_col):

	dx, dy, yi = x1 - x0, y1 - y0, 1

	if dy < 0:
		yi, dy = -1, -dy

	D, y = (2 * dy) - dx, y0

	for x in range(x0, x1 + 1):

		buffer[y][x] = '8;2;%d;%d;%dm' % f_col
		state[y][x] = True 

		if D > 0:
			y = y + yi
			D = D + (2 * (dy - dx))

		else:
			D = D + 2*dy


def rect(buffer, state, x, y, width, height, f_col=(255, 255, 255)):

	for a in range(height):

		buffer[y + a][x] = '8;2;%d;%d;%dm' % f_col
		buffer[y + a][x + width - 1] = '8;2;%d;%d;%dm' % f_col

		state[y + a][x] = True
		state[y + a][x + width - 1] = True

	for b in range(1, width - 1):

		buffer[y][x + b] = '8;2;%d;%d;%dm' % f_col
		buffer[y + height - 1][x + b] = '8;2;%d;%d;%dm' % f_col

		state[y][x + b] = True
		state[y + height - 1][x + b] = True


def circle(buffer, state, cx, cy, rad, f_col=(255, 255, 255)):
		
	x, y, m = 0, rad, 5 - 4 * rad

	while x <= y:

		circle_add(buffer, state, cx, cy, x, y, f_col)

		if m > 0:

			y -= 1
			m -= 8 * y

		x += 1
		m += 8 * x + 4


def circle_add(buffer, state, cx, cy, x, y, f_col):

	buffer[cy + y][cx + x] = '8;2;%d;%d;%dm' % f_col
	buffer[cy + y][cx - x] = '8;2;%d;%d;%dm' % f_col
	buffer[cy - y][cx + x] = '8;2;%d;%d;%dm' % f_col
	buffer[cy - y][cx - x] = '8;2;%d;%d;%dm' % f_col

	buffer[cy + x][cx + y] = '8;2;%d;%d;%dm' % f_col
	buffer[cy + x][cx - y] = '8;2;%d;%d;%dm' % f_col
	buffer[cy - x][cx + y] = '8;2;%d;%d;%dm' % f_col
	buffer[cy - x][cx - y] = '8;2;%d;%d;%dm' % f_col

	state[cy + y][cx + x] = True
	state[cy + y][cx - x] = True
	state[cy - y][cx + x] = True
	state[cy - y][cx - x] = True

	state[cy + x][cx + y] = True
	state[cy + x][cx - y] = True
	state[cy - x][cx + y] = True
	state[cy - x][cx - y] = True


class Image:

	def __init__(self, buffer, state, path):

		self.pos = (0, 0)
		self.size = (8, 8)

		self._buffer = buffer
		self._state = state

		self._img = []
		self._img_size = None

		self._cache = []
		self._cache_size = None

		self.__decode_ppm(path)


	@property
	def pos(self):
		return self._pos

	@pos.setter
	def pos(self, new_pos):
		self._pos = new_pos

	@property
	def size(self):
		return self._size

	@size.setter
	def size(self, new_size):
		self._size = new_size


	def __decode_ppm(self, path):

		with open(path, 'rb') as file:

			if not file.readline().decode('ascii').startswith('P6'):
				raise RuntimeError('Failed to open ', path)

			self._img_size = tuple(map(int, file.readline().decode('ascii').split()))

			if not file.readline().decode('ascii').startswith('255'):
				raise RuntimeError('Failed to open ', path)

			content = file.read(3 * self._img_size[0] * self._img_size[1])
			indice = 0

			for a in range(self._img_size[1]):
				self._img.append([])

				for b in range(self._img_size[0]):

					self._img[-1].append((content[indice], content[indice+1], content[indice+2]))
					indice += 3


	def __resize(self, hide):

		self._cache.clear()
		self._cache_size = self.size

		of_x = self._img_size[0] / self._cache_size[0]
		of_y = self._img_size[1] / self._cache_size[1]

		x, y = 0, 0

		for a in range(self._cache_size[1]):
			for b in range(self._cache_size[0]):

				if self._img[round(y)][round(x)] != hide:
					self._cache.append('8;2;%d;%d;%dm' % self._img[round(y)][round(x)])

				else:
					self._cache.append(None)

				x += of_x

			x = 0
			y += of_y


	def show(self, hide=None):

		force = False
		if self.size != self._cache_size:
			self.__resize(hide)
			force = True

		indice = 0
		for y in range(self.pos[1], self._cache_size[1] + self.pos[1]):
			for x in range(self.pos[0], self._cache_size[0] + self.pos[0]):

				if (self._buffer[y][x] != self._cache[indice] or force) and self._cache[indice] != None:

					self._buffer[y][x] = self._cache[indice]
					self._state[y][x] = True

				indice += 1


import nsii

nsii = nsii.Nsii()       # nsii class initialization

nsii.fps_target = 'max'  # set frame rate target
nsii.p_size = 8          # set pixel size

background = nsii.new_image('image.ppm')  # initializing a new image

while True:

	background.size = nsii.size  # set background size to window size
	background.show()            # show the background

	ox, oy = nsii.m_pos
	while nsii.m_click('left'):  # while the left mouse button is pressed

		x, y = nsii.m_pos  # get mouse coordinates 
		background.show()  # show the background

		if x < ox and y < oy:
			nsii.rect(x, y, ox-x, oy-y)  # draw a rectangle

		elif x < ox:
			nsii.rect(x, oy, ox-x, y-oy)  # draw a rectangle

		elif y < oy:
			nsii.rect(ox, y, x-ox, oy-y)  # draw a rectangle

		else:
			nsii.rect(ox, oy, x-ox, y-oy)  # draw a rectangle

		# display information
		nsii.name = f'framerate : {nsii.fps:.1f} size : {nsii.size} client_size : {nsii.client_size} win_pos : {nsii.pos} m_pos : {nsii.m_pos}'

		nsii.draw()  # draw a new frame

	# display information
	nsii.name = f'framerate : {nsii.fps:.1f} size : {nsii.size} client_size : {nsii.client_size} win_pos : {nsii.pos} m_pos : {nsii.m_pos}'

	nsii.draw()  # draw a new frame